﻿//---------------------
// YureBone.cs
// 　ricopinさんのSpringBoneコンポーネントのソースをかなり参考にしています。
// 　数学的にはricopinさんのものとほぼ変わらない式を使っています。ありがとうございます。
// 　アルゴリズムは同じですが、権利クリアのため、数式表現は変えられるところは全箇所変わってると思います。
// 　YureBone.cs YureCollider.cs YureManager.csに関して私（高橋直樹）は権利を主張しません。
// 改造点：
// 　1.ツールチップヒントをつけた。
// 　2.YureManagerに実行時に自動で子のボーンが登録され、正しい順番で呼び出される（手動登録の必要がない）
// 　3.YureManagerがYureColliderを管理しているのでYureBoneにコライダーを手動登録する必要がない
// 　　（全てのボーンにおいて全てのコライダーとの衝突判定が実行されるので、たとえばSD-Unityちゃんの
// 　　　頭のコライダーは少し大きすぎて、デフォルトのまま置き換えると前髪が当たります。
//     　よって、厳密な意味では互換ではなくなっていますが、動作としてはこちらのほうが自然だと思う。）
// 　4.ボーンの向きを、次のボーンのローカル座標から自動計算するようにした。
// 　5.無駄だと思われるΔt^2の割り算とかけ算を相殺させて消した
// 　6.同じベクトルの足し算を2倍（スカラー倍）に変えたり、一度正規化して長さをかけていたのを
// 　  長さの比をかけるようにしたりなど、数式の記述を多少いじった。計算量は増えてはないと思う。
//---------------------

using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace Yuremono
{ 
	public class YureBone : MonoBehaviour
	{
        [TooltipAttribute("次のボーン")]
        public Transform child;
        [TooltipAttribute("衝突判定用の半径")]
        public float radius=0.05f;
        [TooltipAttribute("ボーンが元の方向に伸びようとする力")]
        public float stiffnessForce = 0.01f;
        [TooltipAttribute("揺れの減衰力")]
        public float dragForce = 0.4f;
        [TooltipAttribute("常にボーンに掛かる力（重力など）")]
        public Vector3 springForce = new Vector3(0.0f, -0.0001f, 0.0f);

        private Vector3 boneAxis;
        private float springLength;//バネの長さ
        private Quaternion localRotation;//初期状態の回転
        private Vector3 currTipPos;//現フレームの次のボーンの位置
        private Vector3 prevTipPos;//前フレームの次のボーンの位置
        private YureManager manager;
        

        private void Awake()
        {
            //ボーンの元々の位置・回転を保存
            localRotation = transform.localRotation;

            //YureManagerを検索。親オブジェクトのどれかにあると想定している。
            Transform t = transform;
            while (true) { 
                var ym = t.GetComponent<YureManager>();
                if (ym == null)
                {
                    t = t.parent;
                    if (t == null) Debug.LogError(gameObject.name+" : YureManagerが見つかりません。");
                }
                else
                {
                    manager = ym;
                    break;
                }
            }
        }

        private void OnDrawGizmos()
        {
            Gizmos.color = Color.yellow;
            Gizmos.DrawWireSphere(currTipPos, radius);
        }

        private void Start ()
        {
            //ボーンの方向ベクトルを設定
            boneAxis = child.transform.localPosition.normalized;
            //ボーンの長さを設定（この長さが保持される）
            springLength = Vector3.Distance(transform.position, child.position);

            //現在位置と次回の位置、二つとも初期位置で初期化
            currTipPos = child.position;
            prevTipPos = child.position;
        }

        public void UpdateSpring()
        {
            //YureManagerから呼ばれる。この更新処理。

            //前回の位置を計算に使うため保存
            Vector3 ppos = prevTipPos;//このフレームで使う「前フレームの位置」x(t-Δt)
            prevTipPos = currTipPos;//現在の位置を保存し次のフレームでの計算に使う


            //次のボーン位置の計算
            //（高橋注：以下、Δt^2の計算を省略したり、同じベクトル同士の足し算を２倍にしたり、弄っています）
            //回転をリセット
            transform.localRotation = localRotation;
            //ボーンの方向に伸びようとする力(Δt^2で割るべきだが後でかけるため相殺)
            Vector3 force = transform.rotation * boneAxis * stiffnessForce;
            //減衰（前フレームの位置に戻ろうとする力）(Δt^2で割るべきだが後でかけるため相殺)
            force += (ppos - currTipPos) * dragForce;// 前回の位置まで引き戻す力を加えている。
            //常にボーンに掛かる力(Δt^2で割るべきだが後でかけるため相殺)
            force += springForce;
            //ベレの方法（検索して調べた。ようは積分のテイラー展開らしい wikipedia等参照）
            currTipPos = 2 * currTipPos - ppos + force;
            //x(t+Δt)=2X(t)-X(t-Δt)+F/M*Δt^2
            //(Mは重さだけどここでは1で考えてるようだ。Δt^2をここでかけるので相殺で消える）
            ResizeBone();//ボーンの長さを元と同じにする
            

            //マネージャからコライダー情報を引き出し衝突判定
            //（高橋注：Managerに自動で登録してそれを使うように変更）
            List<YureCollider> c=manager.GetColliders();
            for (int i = 0; i < c.Count; i++)
            {
                float d=Vector3.Distance(currTipPos, c[i].transform.position);
                float hitd = radius + c[i].radius;
                if (d <= hitd)
                {
                    //球の衝突判定。ぶつかっている場合
                    //高橋注：SpringBoneと同じ結果になる処理だが表現は変えています。（著作権クリアのため）
                    Vector3 v = currTipPos - c[i].transform.position;//衝突したコライダーから遠ざかる方向へのベクトル
                    currTipPos += v * (1.0f - d / hitd);//ちょうどぶつからなくなるところまで動かす      
                    ResizeBone();//ボーンの長さを元と同じにする
                }
            }

            //回転を適用する（高橋注：ここは変えようがないから許して。）
            Vector3 aimVector = transform.TransformDirection(boneAxis);//ボーンの方向ベクトル（ワールド座標系）
            //ボーンの方向ベクトルが今回の正しい方向ベクトルに合うように回転させるクォータニオンを計算
            Quaternion aimRotation = Quaternion.FromToRotation(aimVector, currTipPos - transform.position);
            transform.rotation = aimRotation * transform.rotation;//回転させる。
        }

        private void ResizeBone ()
        {
            //方向は同じで、ボーンの長さを元の長さと同じにする。
            //結果は同じだけど数式表現を変えています（著作権クリアのため）
            Vector3 v = currTipPos - transform.position;
            currTipPos = v * (springLength / v.magnitude) + transform.position;
        }
    }
}